""":mod:`{{ name }}` package

{{ copyright_license_rst }}
"""
import importlib.metadata

try:
    # We only have a version once the package is installed.
    __version__ = importlib.metadata.version("{{ name }}")
except importlib.metadata.PackageNotFoundError:
    pass
